//(c) 2016 Dmitry Grinberg   http://dmitry.gr   me@dmitry.gr   see LICENSE.txt

#include "wincompat.h"
#include <string.h>
#include <stdio.h>
#include "flash.h"
#include "kln.h"

#define MAX_UPDATE_SIZE		0x00300000UL			//3MB


static bool dumpToFile(const char *name, const uint8_t *data, uint32_t len)
{
	bool ret = false;
	
	FILE *out = fopen(name, "wb");
	
	if (!out)
		fprintf(stderr, "Failed to open \"%s\" for write\n", name);
	else {
		
		if (fwrite(data, 1, len, out) != len)
			fprintf(stderr, "Failed to write %lu bytes to \"%s\"\n", (unsigned long)len, name);
		else
			ret = true;
		
		fclose(out);
	}
	
	return ret;
}


int main(int argc, char **argv)
{
	bool needBackup = false, needDecryption = true, needSaveDecrypted = false, needFlash = true, needAllowDiffKeys = false, needUseCardKeys = false;
	static uint8_t encryptedUpdate[MAX_UPDATE_SIZE], decryptedUpdate[MAX_UPDATE_SIZE];
	uint32_t encryptedUpdateSz = 0, decryptedUpdateSz = 0;
	const char *self = argv[0];
	FILE *infile = NULL;
	
	fprintf(stderr, "KLN89B updater tool by Dmitry Grinberg  http://dmitry.gr   me@dmitry.gr\n\n\n");
	
	if (argc < 2)
		goto usage;
	
	argc--;
	argv++;
	
	while(argc) {
		
		if (!strcmp(*argv, "-b")) {
			
			needBackup = true;
			argc--;
			argv++;
		}
		else if (!strcmp(*argv, "-d")) {
			
			needSaveDecrypted = true;
			needFlash = false;
			argc--;
			argv++;
		}
		else if (!strcmp(*argv, "-u")) {
			
			needDecryption = false;
			argc--;
			argv++;
		}
		else if (!strcmp(*argv, "-i")) {
			
			needFlash = false;
			argc--;
			argv++;
		}
		else if (!strcmp(*argv, "-Xu")) {
			
			if (needAllowDiffKeys)
				goto usage;
			needAllowDiffKeys = true;
			argc--;
			argv++;
		}
		else if (!strcmp(*argv, "-Xc")) {
			
			if (needAllowDiffKeys)
				goto usage;
			needAllowDiffKeys = true;
			needUseCardKeys = true;
			argc--;
			argv++;
		}
		else
			break;
	}
	
	if (argc >= 2)
		goto usage;
	
	//do the backup if requested 
	decryptedUpdateSz = sizeof(decryptedUpdate);
	if (needBackup && (!flashMemoryRead(decryptedUpdate, &decryptedUpdateSz) || !dumpToFile("backup.bin", decryptedUpdate, decryptedUpdateSz))) {
		
		fprintf(stderr, "Backup failed. Aborting\n");
		goto out;
	}
	
	//all remaining operations only work if we have a file
	if (!argc)
		goto out_success;
	
	//if we need a file, make sure we can get it
	if (strcmp(argv[0], "-")) {
		
		infile = fopen(argv[0], "rb");
		if (!infile) {
			
			fprintf (stderr, "Cannot open \"%s\"\n", argv[0]);
			goto out;
		}
	}
	else {
		infile = stdin;
		if (isatty(fileno(stdin))) {
			
			fprintf (stderr, "You forgot to redirect a pipe into this process. Refusing to read from console stdin\n");
			goto out;
		}
	}
	
	//read the file in
	for (encryptedUpdateSz = 0; encryptedUpdateSz < MAX_UPDATE_SIZE; encryptedUpdateSz++) {
		
		int c = fgetc(infile);
		if (c == EOF)
			break;
		encryptedUpdate[encryptedUpdateSz] = c;
	}
	if (!feof(infile)) {
		
		fprintf(stderr, "Input file is too large (max allowed size: %lu bytes)\n", MAX_UPDATE_SIZE);
		goto out;
	}
	fprintf(stderr, "Read %lu-byte input file\n", (unsigned long)encryptedUpdateSz);
	
	//decrypt file if it needs it
	decryptedUpdateSz = sizeof(decryptedUpdate);
	if (!needDecryption) {
		
		decryptedUpdateSz = encryptedUpdateSz;
		memcpy(decryptedUpdate, encryptedUpdate, encryptedUpdateSz);
	}
	else if (needDecryption && !klnDecrypt(decryptedUpdate, &decryptedUpdateSz, encryptedUpdate, encryptedUpdateSz)) {
			
		fprintf(stderr, "Error during decryption. Aborting\n");
		goto out;
	}
	
	//perform checksum verification & show info
	if (!klnVerifyChecksumsAndShowInfo(decryptedUpdate, decryptedUpdateSz)) {
		
		fprintf(stderr, "Decrypted file has checksum errors. Aborting\n");
		goto out;
	}
	
	//if we were asked to write out the decrypted file, do so
	if (needSaveDecrypted && !dumpToFile("update.bin", decryptedUpdate, decryptedUpdateSz)) {
		
		fprintf(stderr, "Failed to write decrypted update. Aborting\n");
		goto out;
	}

	//flash if requested
	if (needFlash && !flashMemoryWrite(decryptedUpdate, decryptedUpdateSz, needAllowDiffKeys, needUseCardKeys)) {
		
		fprintf(stderr, "Failed to write update to card. Aborting\n");
		goto out;
	}
	
out_success:
	fprintf(stderr, "All operations succeeded\n");

out:
	if (infile && infile != stdin)
		fclose(infile);
	flashDeinit();
	if (isatty(fileno(stdin))) {
		fprintf(stderr, "press ENTER to quit\n");
		getchar();
	}
	return 0;

usage:
	fprintf(stderr,	"USAGE: %s [OPTIONS] database.dat\n"
					"  If no  options are given, device is updated with file passed in silently\n"
					"  Options:\n"
					"   -b = backup current card to \"backup.bin\"\n"
					"   -d = decrypt update only to \"update.bin\"\n"
					"   -u = database path given is not encrypted (eg: a previous backup)\n"
					"   -i = show update info only\n"
					"   -Xu = allow write of update with mismatching keys. Use update's keys\n"
					"   -Xc = allow write of update with mismatching keys. Use card's keys\n"
					"   \"-\" can be used in lieu of filename to use stdin\n"
					"   Obviously -Xu and -Xc cannot be used at the same time\n",
					self);
	goto out;
}